#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/poll.h>
#include <fcntl.h>
#include <unistd.h>
#include <limits.h>
#include <errno.h>
#include <string.h>
#include <ctype.h>

#define MAX_LENGTH 256

void get_passlen(char *arg, short int *len);
int give_password(void *buf, short int len);
char select_char(int num, const short int *flag);
void generate_password(char *word, short int len, const short int *flag);
void set_flags(int num, char **keys, short int *flag);
void show_hint(void);

int main(int argc, char **argv)
{
	short int passlen, flag[4];
	char password[MAX_LENGTH+1];

	if (argc < 2) {
		printf("Too few keys. Run the application with ? key to get help\n");
		exit(1);
	}

	get_passlen(argv[1], &passlen);
	set_flags(argc, argv, flag);
	generate_password(password, passlen, flag);
	give_password(password, passlen);

	return 0;	
}

// function gets the length
void get_passlen(char *arg, short int *len)
{
	if (arg[0] == '?') {
		show_hint();
	} else {
		*len = atoi(arg);

		// if user entered any trash
		if (*len  == 0) {
			printf("The first key must contain the length as integer positive number\n");
			printf("Run the application with ? key to get help");
			exit(3);
		}
	
		// if length exceeds maximal value
		if (*len > (MAX_LENGTH + 1) ) {
			printf("You entered too big length\n");
			exit(4);
		}
	}
}

// function returns a character
// see ascii table to understand it
char select_char(int num, const short int *flag)
{
	// if the number is negative we return '\0'.
	// so, we'll do it when the number isn't according with flags risen
	if (num < 0)
		return '\0';
	
	// define the 1st special symbols interval
	// and check according flag
	else if (num < 15) {
		if (flag[0] == 1)
			return (char)(num + 33);
		else
			return '\0';
	}
	
	// define numbers interval
	// and check according flag
	else if (num < 25) {
		if (flag[3] == 1)
			return (char)(num + 33);
		else
			return '\0';
	}

	// define the 2nd special symbols interval
	// and check according flag
	else if (num < 32) {
		if (flag[0] == 1)
			return (char)(num + 33);
		else
			return '\0';
	}
	// define upper case letters interval
	// and check according flag	
	else if (num < 58) {
		if (flag[1] == 1)
			return (char)(num + 33);
		else
			return '\0';
	}

	// define the 3d special symbols interval
	// and check according flag
	else if (num < 64) {
		if (flag[0] == 1)
			return (char)(num + 33);
		else
			return '\0';
	}
	
	// define lower case letters interval
	// and check according flag
	else if (num < 90) {
		if (flag[2] == 1)
			return (char)(num + 33);
		else
			return '\0';
	}

	// define the 4th special symbols interval
	// and check according flag
	else if (num < 94) {
		if (flag[0] == 1)
			return (char)(num + 33);
		else
			return '\0';
	}
}

// function generating password
void generate_password(char *word, short int len, const short int *flag)
{
	int fd, i, res = 0;
	unsigned short seeding[3];
	char newchar;
	struct pollfd pfd;

	// /dev/random - random number generator, read the random(4)
	fd = open("/dev/random", O_RDONLY);
	if (fd == -1) {
		printf("Can't initialize the random\n");
		exit(5);
	}

	pfd.fd = fd;
	pfd.events = POLLIN;
	pfd.revents = 0;

	//fill seeding with random bits from /dev/random.
	//if we have too few bits on /dev/random, we ask
	//user to do somehing to fill /dev/random
	while (res != 3) {
		res = read(fd, seeding, 3);
		if (res == -1) {
			printf("Can't extract the seeding\n");
			exit(6);
		}
		if (res < 3) {
			printf("Please move mouse, run any other application or listen to the music.");
			i = poll(&pfd, 1, 7000);
			if (i == -1) {
				if (errno == EINTR)
					printf("Poll broken by signal\n");
				else
					printf("Poll mistaked\n");
			}
		}
		
	}

	//collect the seeding - lrand48(3) returns random numbers
	seed48(seeding);

	//fill each character password with random character
	for (i = 0; i < len; i++) {
		do {
			res = lrand48() % 94;
			newchar = select_char(res, flag);
		} while (newchar == '\0');
		word[i] = newchar;
	}

	close(fd);
}

// function sending password to stdout
int give_password(void *buf, short int len) 
{
	int res, left;
	left = len;
	
	while (left) {
		res = fwrite(buf, sizeof(char), left, stdout);
		if(res == -1)  {
			if (errno == EINTR)
				continue;
			printf("Can't give the password\n");
			exit(7);
		}
		buf += res;
		left -= res;
	}
	return 0;
}

// function setting the flags for password generating.
// flag[0] is associated with special symbols
// flag[1] - with upper case latin symbols
// flag[2] - with lower case latin symbols
// flag[3] - with numbers.
void set_flags(int num, char **keys, short int *flag)
{
	int i;

	if (num == 2) {
		flag[0] = 1;
		flag[1] = 1;
		flag[2] = 1;
		flag[3] = 1;
	} else if (num < 7) {

		for (i = 2; i < num; i++) {
			switch (keys[i][0]) {
			case 's':
				flag[0] = 1;
				break;
			case 'u':
				flag[1]= 1;
				break;
			case 'l':
				flag[2]= 1;
				break;
			case 'n':
				flag[3]= 1;
				break;
			default:
				printf("Please check entered keys\n");
				printf("Run the application with ? key to get help\n");
				exit(8);
			}
		}
	} else {
		printf("Too many keys\n");
		exit(9);
	}
}
// function showing hint for user
void show_hint(void)
{
	printf("\nSyntax: passgen [length] [key] ...\n");
	printf("[length] - length of the password you want to generate as integer positive number smaller than 256.\n");
	printf("[key] - one of keys you can see in the table below. You can use some keys,\n");
	printf("Key\tMeaning\n");
	printf(" s\tspecial symbol\n");
	printf(" u\tupper case symbol\n");
	printf(" l\tlower case symbol\n");
	printf(" n\tnumber from 0 to 9\n");
	exit(2);
}
